class JSONRPCException(Exception):
    CODE = None
    MESSAGE = None

    def __init__(self, id: str = None, errors: dict = None) -> None:
        self._id = id
        self._errors = errors or {}

    @property
    def id(self):
        return self._id

    @property
    def errors(self):
        return self._errors


class JSONRPCMethodNotFoundException(JSONRPCException):
    CODE = -32601
    MESSAGE = "Method not found."


class JSONRPCInvalidParamsException(JSONRPCException):
    CODE = -32602
    MESSAGE = "Invalid params."


class JSONRPCInvalidRequestException(JSONRPCException):
    CODE = -32600
    MESSAGE = "Invalid Request."


class JSONRPCParseErrorException(JSONRPCException):
    CODE = -32700
    MESSAGE = "Parse error."



from tests import client


def test_batch():
    payload = [
        {"jsonrpc": "2.0", "method": "sum", "params": [1, 2, 4], "id": "1"},
        {"jsonrpc": "2.0", "method": "notify_hello", "params": [7]},
    ]
    # payload = [
    #     {"jsonrpc": "2.0", "method": "sum", "params": [1, 2, 4], "id": "1"},
    #     {"jsonrpc": "2.0", "method": "notify_hello", "params": [7]},
    #     {"jsonrpc": "2.0", "method": "subtract_positional", "params": [42, 23], "id": "2"}, good
    #     {"foo": "boo"}, bad
    #     {"jsonrpc": "2.0", "method": "foo.get", "params": {"name": "myself"}, "id": "5"}, good
    #     {"jsonrpc": "2.0", "method": "get_data", "id": "9"} good
    # ]
    response = client.post("/api/", json=payload)
    assert response.json() == [
        {"jsonrpc": "2.0", "result": {"result": 7}, "id": "1"},
        {},
    ]
    # assert response.json() == [
    #     {"jsonrpc": "2.0", "result": {"result": 7}, "id": "1"},
    #     {},
    #     {"jsonrpc": "2.0", "result": {'result': 19}, "id": "2"},
    #     {"jsonrpc": "2.0", "error": {"code": -32600, "message": "Invalid Request"}, "id": None},
    #     {"jsonrpc": "2.0", "error": {"code": -32601, "message": "Method not found.", "data": {}}, "id": "5"},
    #     {"jsonrpc": "2.0", "result": {"result": ["hello", 5]}, "id": "9"}
    # ]



from . import client


# JSON


def test_payload_as_empty_dict():
    payload = {}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "None",
        "error": {"code": -32600, "message": "Invalid Request.", "data": {}},
    }


def test_payload_as_empty_list():
    payload = []
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "None",
        "error": {"code": -32600, "message": "Invalid Request.", "data": {}},
    }


def test_incorrect_payload():
    payload = [1]
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "None",
        "error": {"code": -32600, "message": "Invalid Request.", "data": {}},
    }


# PARAMS


def test_positional_parameters():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract_positional",
        "params": [42, 23],
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "1", "result": 19}


def test_positional_parameters_2():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract_positional",
        "params": [23, 42],
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "1", "result": -19}


def test_named_parameters():
    payload = {
        "jsonrpc": "2.0",
        "method": "SubtractMethod",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "1", "result": 19}


def test_named_parameters_2():
    payload = {
        "jsonrpc": "2.0",
        "method": "SubtractMethod",
        "params": {"y": 23, "x": 42},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "1", "result": 19}


def test_named_parameters_3():
    payload = {
        "jsonrpc": "2.0",
        "method": "sum",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "1", "result": {"sum": 65}}


def test_params_not_object():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": "", "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"params": "Did not match any valid type."},
        },
    }


def test_params_as_invalid_object():
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"params": "Required param: 'x'"},
        },
    }


def test_params_as_invalid_list():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract_positional",
        "params": [1],
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {
                "params": "subtract_positional() missing 1 required positional argument: 'y'"
            },
        },
    }


def test_without_params():
    payload = {"jsonrpc": "2.0", "method": "my_method", "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.status_code == 200


# ID


def test_id_as_integer():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": 1,
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": 1, "result": 19}


def test_id_as_string():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "abc",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": "abc", "result": 19}


def test_id_as_null():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": None,
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": None, "result": 19}


def test_empty_id():
    payload = {
        "jsonrpc": "2.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {"jsonrpc": "2.0", "id": None, "result": 19}


def test_notification():
    """
    Notification
    """
    payload = {"jsonrpc": "2.0", "method": "subtract", "params": {"x": 42, "y": 23}}
    response = client.post("/api/", json=payload)
    assert response.json() == {}


# JSONRPC


def test_jsonrpc_as_integer():
    payload = {
        "jsonrpc": 2,
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "Must be a string."},
        },
    }


def test_empty_jsonrpc():
    payload = {
        "jsonrpc": "",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "Must not be blank."},
        },
    }


def test_jsonrpc_wrong_value():
    payload = {
        "jsonrpc": "3.0",
        "method": "subtract",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "Must match the pattern /2.0/."},
        },
    }


def test_without_jsonrpc():
    payload = {"method": "subtract", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"jsonrpc": "This field is required."},
        },
    }


# METHOD


def test_not_registered_method():
    payload = {
        "jsonrpc": "2.0",
        "method": "non_existing_method",
        "params": {"x": 42, "y": 23},
        "id": "1",
    }
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {"code": -32601, "message": "Method not found.", "data": {}},
    }


def test_without_method():
    payload = {"jsonrpc": "2.0", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"method": "This field is required."},
        },
    }


def test_with_empty_method():
    payload = {"jsonrpc": "2.0", "method": "", "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"method": "Must not be blank."},
        },
    }


def test_method_as_integer():
    payload = {"jsonrpc": "2.0", "method": 1, "params": {"x": 42, "y": 23}, "id": "1"}
    response = client.post("/api/", json=payload)
    assert response.json() == {
        "jsonrpc": "2.0",
        "id": "1",
        "error": {
            "code": -32602,
            "message": "Invalid params.",
            "data": {"method": "Must be a string."},
        },
    }


# def test_with_method_name_starting_with_rpc_period():
#     pass



import types

from starlette_jsonrpc import dispatcher


def test_adding_methods_to_dispatcher():
    func = dispatcher.methods_map.get("subtract")
    assert func is not None
    assert isinstance(func, types.FunctionType) is True

    func = dispatcher.methods_map.get("SubtractMethod")
    assert func is not None
    assert isinstance(func, types.FunctionType) is True

    func = dispatcher.methods_map.get("subtract_positional")
    assert func is not None
    assert isinstance(func, types.FunctionType) is True



from json import JSONDecodeError

from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.responses import Response

from starlette_jsonrpc import dispatcher
from starlette_jsonrpc.constants import JSONRPC_VERSION
from starlette_jsonrpc.exceptions import JSONRPCException
from starlette_jsonrpc.exceptions import JSONRPCInvalidParamsException
from starlette_jsonrpc.exceptions import JSONRPCInvalidRequestException
from starlette_jsonrpc.exceptions import JSONRPCMethodNotFoundException
from starlette_jsonrpc.exceptions import JSONRPCParseErrorException
from starlette_jsonrpc.schemas import JSONRPCErrorResponse
from starlette_jsonrpc.schemas import JSONRPCRequest
from starlette_jsonrpc.schemas import JSONRPCResponse


class JSONRPCEndpoint(HTTPEndpoint):
    async def post(self, request: Request) -> Response:
        try:
            response = await self._get_response(request)
        except JSONRPCInvalidParamsException as exc:
            return self._get_exception_response(exc)
        except JSONRPCMethodNotFoundException as exc:
            return self._get_exception_response(exc)
        except JSONRPCInvalidRequestException as exc:
            return self._get_exception_response(exc)
        except JSONRPCParseErrorException as exc:
            return self._get_exception_response(exc)

        return JSONResponse(response)

    async def _get_response(self, request: Request) -> dict:
        try:
            req = await request.json()
        except JSONDecodeError:
            raise JSONRPCParseErrorException()
        except Exception:
            raise JSONRPCInvalidRequestException()

        if not self._valid_request(req):
            raise JSONRPCInvalidRequestException()

        if not req:
            raise JSONRPCInvalidRequestException()

        if self._is_notification(req):
            return {}

        data, errors = JSONRPCRequest.validate_or_error(req)
        id = req.get("id")

        if errors:
            raise JSONRPCInvalidParamsException(id, errors)

        method = data.get("method")
        func = dispatcher.methods_map.get(method)

        if not func:
            raise JSONRPCMethodNotFoundException(id)

        params = data.get("params")

        result = await self._get_result(params, func, id)

        response = JSONRPCResponse.validate(
            {"id": id, "jsonrpc": JSONRPC_VERSION, "result": result}
        )
        return dict(response)

    @staticmethod
    def _valid_request(req) -> bool:
        if isinstance(req, dict):
            return True
        if isinstance(req, list):
            if all([isinstance(elem, dict) for elem in req]):
                return True
        return False

    @staticmethod
    def _is_notification(req: dict) -> bool:
        if all(k in req for k in ("jsonrpc", "method", "params")) and not "id" in req:
            return True
        return False

    @staticmethod
    async def _get_result(params, func, id: str = None) -> dict:
        if isinstance(params, list):
            try:
                return await func(*params)
            except TypeError as e:
                errors = {"params": f"{e}"}
                raise JSONRPCInvalidParamsException(id, errors)
        elif isinstance(params, dict):
            try:
                return await func(params)
            except KeyError as e:
                errors = {"params": f"Required param: {e}"}
                raise JSONRPCInvalidParamsException(id, errors)
        else:
            raise JSONRPCInvalidRequestException()

    @staticmethod
    def _get_exception_response(exc: JSONRPCException) -> JSONResponse:
        response = JSONRPCErrorResponse.validate(
            {
                "jsonrpc": "2.0",
                "id": str(exc.id),
                "error": {
                    "code": exc.CODE,
                    "message": exc.MESSAGE,
                    "data": {key: value for (key, value) in exc.errors.items()},
                },
            }
        )
        return JSONResponse(dict(response))
